@@ -16,10 +16,13 @@ use FindBin;
 
 our $ACTIVEMQ_VERSION = '5.2.0';
 
-our @EXPORT = qw/ start_server /;
+our @EXPORT = qw/ start_server check_amq_broker /;
+
+my $mq;
+
+sub check_amq_broker {
+    my ($stomp);
 
-sub start_server {
-    my ($mq, $stomp);
     eval {
         $stomp = Net::Stomp->new( { hostname => 'localhost', port => 61613 } );
     };
@@ -30,7 +33,7 @@ sub start_server {
             exit;
         }
 
-        $mq = Alien::ActiveMQ->run_server($ACTIVEMQ_VERSION);
+        $mq ||= Alien::ActiveMQ->run_server($ACTIVEMQ_VERSION);
 
         eval {
             $stomp = Net::Stomp->new( { hostname => 'localhost', port => 61613 } );
@@ -41,6 +44,12 @@ sub start_server {
         }
     }
 
+    return $stomp;
+}
+
+sub start_server {
+    my $stomp = check_amq_broker();
+
     $SIG{CHLD} = 'IGNORE';
     unless (fork()) {
         my $libs = join(' ', map { "-I$_" } @INC);
@@ -54,11 +63,12 @@ sub start_server {
 
         exit 0;
     }
-    print STDERR "server started, waiting for spinup...";
-    sleep 20;
+    diag "server started, waiting for spinup...";
+    sleep($ENV{NET_STOMP_DELAY}||20);
 
     $stomp->{___activemq} = $mq if $mq;
     StompRole->meta->apply($stomp);
     return $stomp;
 }
 
+1;
